from typing import Optional, List

from secoda_common.models.table import Badge

import attr
from marshmallow_annotations.ext.attrs import AttrsSchema


@attr.s(auto_attribs=True, kw_only=True)
class LineageItem:
    key: str  # down/upstream table/col/task key
    level: int  # upstream/downstream distance from current resource
    description: Optional[str]
    badges: Optional[List[Badge]] = None
    usage: Optional[int] = None  # statistic to sort lineage items by


class LineageItemSchema(AttrsSchema):
    class Meta:
        target = LineageItem
        register_as_scheme = True


@attr.s(auto_attribs=True, kw_only=True)
class Lineage:
    key: str  # current table/col/task key
    direction: str  # upstream/downstream/both
    depth: int  # how many levels up/down 0 == all
    upstream_entities: List[LineageItem]  # list of upstream entities
    downstream_entities: List[LineageItem]  # list of downstream entities


class LineageSchema(AttrsSchema):
    class Meta:
        target = Lineage
        register_as_scheme = True


from typing import Optional, List, Dict
from secoda_common.models.user import User
from secoda_common.models.table import TableSummary
from secoda_common.models.dashboard import DashboardSummary
from secoda_common.models.tag import Tag

import attr
from marshmallow_annotations.ext.attrs import AttrsSchema

@attr.s(auto_attribs=True, kw_only=True)
class DictionaryTerm:
  key: str
  name: Optional[str]
  definition: Optional[str] = None
  sql: Optional[str] = None
  owner: Optional[User] = None
  created_at: Optional[int] = None
  updated_at: Optional[int] = None
  tags: Optional[List[Tag]] = None
  related_tables: Optional[List[TableSummary]] = None
  related_dashboards: Optional[List[DashboardSummary]] = None


class DictionaryTermSchema(AttrsSchema):
  class Meta:
    target = DictionaryTerm
    register_as_scheme = True


from typing import Optional, List

import attr
from marshmallow_annotations.ext.attrs import AttrsSchema


@attr.s(auto_attribs=True, kw_only=True)
class PiiTable:
    database: str = attr.ib()
    cluster: str = attr.ib()
    schema: str = attr.ib()
    name: str = attr.ib()
    columns: List[str] = attr.ib()


class PiiTableSchema(AttrsSchema):
    class Meta:
        target = PiiTable
        register_as_scheme = True




import textwrap

# Specifying default mapping for elasticsearch index
# Documentation: https://www.elastic.co/guide/en/elasticsearch/reference/current/mapping.html
# Setting type to "text" for all fields that would be used in search
# Using Simple Analyzer to convert all text into search terms
# https://www.elastic.co/guide/en/elasticsearch/reference/current/analysis-simple-analyzer.html
# Standard Analyzer is used for all text fields that don't explicitly specify an analyzer
# https://www.elastic.co/guide/en/elasticsearch/reference/current/analysis-standard-analyzer.html
TABLE_INDEX_MAP = textwrap.dedent(
    """
    {
    "mappings":{
        "table":{
          "properties": {
            "name": {
              "type":"text",
              "analyzer": "simple",
              "fields": {
                "raw": {
                  "type": "keyword"
                }
              }
            },
            "schema": {
              "type":"text",
              "analyzer": "simple",
              "fields": {
                "raw": {
                  "type": "keyword"
                }
              }
            },
            "display_name": {
              "type": "keyword"
            },
            "last_updated_timestamp": {
              "type": "date",
              "format": "epoch_second"
            },
            "description": {
              "type": "text",
              "analyzer": "simple"
            },
            "column_names": {
              "type":"text",
              "analyzer": "simple",
              "fields": {
                "raw": {
                  "type": "keyword"
                }
              }
            },
            "column_descriptions": {
              "type": "text",
              "analyzer": "simple"
            },
            "tags": {
              "type": "keyword"
            },
            "badges": {
              "type": "keyword"
            },
            "cluster": {
              "type": "text",
              "analyzer": "simple",
              "fields": {
                "raw": {
                  "type": "keyword"
                }
              }
            },
            "database": {
              "type": "text",
              "analyzer": "simple",
              "fields": {
                "raw": {
                  "type": "keyword"
                }
              }
            },
            "key": {
              "type": "keyword"
            },
            "total_usage":{
              "type": "long"
            },
            "unique_usage": {
              "type": "long"
            }
          }
        }
      }
    }
    """
)

DASHBOARD_ELASTICSEARCH_INDEX_MAPPING = textwrap.dedent(
    """
    {
        "settings": {
          "analysis": {
            "normalizer": {
              "lowercase_normalizer": {
                "type": "custom",
                "char_filter": [],
                "filter": ["lowercase", "asciifolding"]
              }
            }
          }
        },
        "mappings":{
            "dashboard":{
              "properties": {
                "group_name": {
                  "type":"text",
                  "analyzer": "simple",
                  "fields": {
                    "raw": {
                      "type": "keyword",
                      "normalizer": "lowercase_normalizer"
                    }
                  }
                },
                "name": {
                  "type":"text",
                  "analyzer": "simple",
                  "fields": {
                    "raw": {
                      "type": "keyword",
                      "normalizer": "lowercase_normalizer"
                    }
                  }
                },
                "description": {
                  "type":"text",
                  "analyzer": "simple",
                  "fields": {
                    "raw": {
                      "type": "keyword"
                    }
                  }
                },
                "group_description": {
                  "type":"text",
                  "analyzer": "simple",
                  "fields": {
                    "raw": {
                      "type": "keyword"
                    }
                  }
                },
                "query_names": {
                  "type":"text",
                  "analyzer": "simple",
                  "fields": {
                    "raw": {
                      "type": "keyword"
                    }
                  }
                },
                "chart_names": {
                  "type":"text",
                  "analyzer": "simple",
                  "fields": {
                    "raw": {
                      "type": "keyword"
                    }
                  }
                },
                "tags": {
                  "type": "keyword"
                },
                "badges": {
                  "type": "keyword"
                }
              }
            }
          }
        }
    """
)

USER_INDEX_MAP = textwrap.dedent("""
{
"mappings":{
    "user":{
      "properties": {
        "email": {
          "type":"text",
          "analyzer": "simple",
          "fields": {
            "raw": {
              "type": "keyword"
            }
          }
        },
        "first_name": {
          "type":"text",
          "analyzer": "simple",
          "fields": {
            "raw": {
              "type": "keyword"
            }
          }
        },
        "last_name": {
          "type":"text",
          "analyzer": "simple",
          "fields": {
            "raw": {
              "type": "keyword"
            }
          }
        },
        "name": {
          "type":"text",
          "analyzer": "simple",
          "fields": {
            "raw": {
              "type": "keyword"
            }
          }
        },
        "total_read":{
          "type": "long"
        },
        "total_own": {
          "type": "long"
        },
        "total_follow": {
          "type": "long"
        }
      }
    }
  }
}""")

JOB_ELASTICSEARCH_INDEX_MAPPING = textwrap.dedent(
    """
    {
        "settings": {
          "analysis": {
            "normalizer": {
              "lowercase_normalizer": {
                "type": "custom",
                "char_filter": [],
                "filter": ["lowercase", "asciifolding"]
              }
            }
          }
        },
        "mappings":{
            "job":{
              "properties": {
                "name": {
                  "type":"text",
                  "analyzer": "simple",
                  "fields": {
                    "raw": {
                      "type": "keyword",
                      "normalizer": "lowercase_normalizer"
                    }
                  }
                },
                "product": {
                  "type":"text",
                  "analyzer": "simple",
                  "fields": {
                    "raw": {
                      "type": "keyword",
                      "normalizer": "lowercase_normalizer"
                    }
                  }
                },
                "description": {
                  "type":"text",
                  "analyzer": "simple",
                  "fields": {
                    "raw": {
                      "type": "keyword"
                    }
                  }
                },
                "tags": {
                  "type": "keyword"
                },
              }
            }
          }
        }
    """
)



from typing import List, Optional

import attr
from marshmallow_annotations.ext.attrs import AttrsSchema
from secoda_common.models.tag import Tag


@attr.s(auto_attribs=True, kw_only=True)
class Badge:
    badge_name: str = attr.ib()
    category: str = attr.ib()


class BadgeSchema(AttrsSchema):
    class Meta:
        target = Badge
        register_as_scheme = True


@attr.s(auto_attribs=True, kw_only=True)
class DashboardSummary:
    uri: str = attr.ib()
    cluster: str = attr.ib()
    group_name: str = attr.ib()
    group_url: str = attr.ib()
    product: str = attr.ib()
    name: str = attr.ib()
    url: str = attr.ib()
    description: Optional[str] = None
    last_successful_run_timestamp: Optional[int] = None
    chart_names: Optional[List[str]] = []
    badges: Optional[List[Badge]] = []
    tags: Optional[List[Tag]] = []


class DashboardSummarySchema(AttrsSchema):
    class Meta:
        target = DashboardSummary
        register_as_scheme = True



from typing import Optional

import attr
from marshmallow_annotations.ext.attrs import AttrsSchema


@attr.s(auto_attribs=True, kw_only=True)
class Query:
    key: str
    query_text: str
    frequency: Optional[int] = None
    description: Optional[str] = None
    last_updated_timestamp: Optional[int] = None


class QuerySchema(AttrsSchema):
    class Meta:
        target = Query
        register_as_scheme = True



from typing import Any, Dict, List, Optional

import attr
from marshmallow import ValidationError, pre_load, validates_schema
from marshmallow_annotations.ext.attrs import AttrsSchema

"""
TODO: Explore all internationalization use cases and
redesign how User handles names

TODO - Delete pre processing of the Data
Once all of the upstream services provide a complete User object we will no
longer need to supplement the User objects as done in `preprocess_data`
"""


@attr.s(auto_attribs=True, kw_only=True)
class User:
    user_id: Optional[str] = None
    email: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    full_name: Optional[str] = None
    display_name: Optional[str] = None
    is_active: bool = True
    github_username: Optional[str] = None
    team_name: Optional[str] = None
    slack_id: Optional[str] = None
    employee_type: Optional[str] = None
    manager_fullname: Optional[str] = None
    manager_email: Optional[str] = None
    manager_id: Optional[str] = None
    role_name: Optional[str] = None
    profile_url: Optional[str] = None
    profile_picture: Optional[str] = None
    background_picture: Optional[str] = None
    role: Optional[str] = None
    dismissed: Optional[List[str]] = None
    has_environment: Optional[bool] = None
    service_account: Optional[bool] = None
    other_key_values: Optional[Dict[str, str]] = attr.ib(factory=dict)  # type: ignore


class UserSchema(AttrsSchema):
    class Meta:
        target = User
        register_as_scheme = True

    # noinspection PyMethodMayBeStatic
    def _str_no_value(self, s: Optional[str]) -> bool:
        # Returns True if the given string is None or empty
        if not s:
            return True
        if len(s.strip()) == 0:
            return True
        return False

    @pre_load
    def preprocess_data(self, data: Dict[str, Any]) -> Dict[str, Any]:
        if self._str_no_value(data.get("user_id")):
            data["user_id"] = data.get("email")

        if self._str_no_value(data.get("profile_url")):
            data["profile_url"] = ""
            if data.get("GET_PROFILE_URL"):
                data["profile_url"] = data.get("GET_PROFILE_URL")(data["user_id"])  # type: ignore

        first_name = data.get("first_name")
        last_name = data.get("last_name")

        if self._str_no_value(data.get("full_name")) and first_name and last_name:
            data["full_name"] = f"{first_name} {last_name}"

        if self._str_no_value(data.get("display_name")):
            if self._str_no_value(data.get("full_name")):
                data["display_name"] = data.get("email")
            else:
                data["display_name"] = data.get("full_name")

        return data

    @validates_schema
    def validate_user(self, data: Dict[str, Any]) -> None:
        if self._str_no_value(data.get("display_name")):
            raise ValidationError(
                '"display_name", "full_name", or "email" must be provided'
            )

        if self._str_no_value(data.get("user_id")):
            raise ValidationError('"user_id" or "email" must be provided')



import os


def req(key):
    if key not in os.environ:
        raise Exception(f'Missing required environment property {key}!')

    return os.environ[key]


def opt(key, default):
    if key not in os.environ:
        print(f'Missing environment property <{key}> --- Using default value "{default}"')
    return os.getenv(key, default)



import functools
import json
import logging
import socket
from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Dict

from flask import current_app as flask_app
from secoda_common.log import action_log_callback
from secoda_common.log.action_log_model import ActionLogParams

LOGGER = logging.getLogger(__name__)
EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)  # use POSIX epoch

# CONFIG KEY FOR caller_retrieval instance
CALLER_RETRIEVAL_INSTANCE_KEY = "CALLER_RETRIEVAL_INSTANCE"


def action_logging(f: Callable[..., Any]) -> Any:
    """
    Decorates function to execute function at the same time triggering action logger callbacks.
    It will call action logger callbacks twice, one for pre-execution and the other one for post-execution.
    Action logger will be called with ActionLogParams

    :param f: function instance
    :return: wrapped function
    """

    @functools.wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        """
        An wrapper for api functions. It creates ActionLogParams based on the function name, positional arguments,
        and keyword arguments.

        :param args: A passthrough positional arguments.
        :param kwargs: A passthrough keyword argument
        """
        metrics = _build_metrics(f.__name__, *args, **kwargs)
        action_log_callback.on_pre_execution(ActionLogParams(**metrics))
        output = None
        try:
            output = f(*args, **kwargs)
            return output
        except Exception as e:
            metrics["error"] = e
            raise e
        finally:
            metrics["end_epoch_ms"] = get_epoch_millisec()
            try:
                metrics["output"] = json.dumps(output)
            except Exception:
                metrics["output"] = output

            action_log_callback.on_post_execution(ActionLogParams(**metrics))

    if LOGGER.isEnabledFor(logging.DEBUG):
        LOGGER.debug("action has been logged")

    return wrapper


def get_epoch_millisec() -> int:
    return (datetime.now(timezone.utc) - EPOCH) // timedelta(milliseconds=1)


def _build_metrics(func_name: str, *args: Any, **kwargs: Any) -> Dict[str, Any]:
    """
    Builds metrics dict from function args
    :param func_name:
    :param args:
    :param kwargs:
    :return: Dict that matches ActionLogParams variable
    """

    metrics = {
        "command": kwargs.get("command", func_name),
        "start_epoch_ms": get_epoch_millisec(),
        "host_name": socket.gethostname(),
        "pos_args_json": json.dumps(args),
        "keyword_args_json": json.dumps(kwargs),
    }  # type: Dict[str, Any]

    caller_retriever = flask_app.config.get(CALLER_RETRIEVAL_INSTANCE_KEY, "")
    if caller_retriever:
        metrics["user"] = caller_retriever.get_caller()
    else:
        metrics["user"] = "UNKNOWN"

    return metrics



import attr

from typing import List, Optional
from secoda_common.models.tag import Tag
from marshmallow_annotations.ext.attrs import AttrsSchema


@attr.s(auto_attribs=True, kw_only=True)
class JobSummary:
    uri: str = attr.ib()
    product: str = attr.ib()
    name: str = attr.ib()
    url: Optional[str] = None
    description: Optional[str] = None
    last_successful_run_timestamp: Optional[int] = None


class JobSummarySchema(AttrsSchema):
    class Meta:
        target = JobSummary
        register_as_scheme = True



import attr

from typing import List, Optional
from marshmallow_annotations.ext.attrs import AttrsSchema

@attr.s(auto_attribs=True, kw_only=True)
class Tag:
    key: str
    tag_name: str
    tag_type: Optional[str] = None
    icon: Optional[str] = None
    description: Optional[str] = None
    last_updated_timestamp: Optional[int] = None
    tag_count: Optional[int] = None


class TagSchema(AttrsSchema):
    class Meta:
        target = Tag
        register_as_scheme = True


from setuptools import find_packages, setup

setup(
    name="secoda-common",
    version="0.1.53",
    description="Common code library for Secoda",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/secoda/common",
    maintainer="Secoda Inc.",
    maintainer_email="andrew.mcewen96@gmail.com",
    packages=find_packages(exclude=["tests*"]),
    dependency_links=[
        (
            "git+https://www.github.com/hilearn/marshmallow-annotations.git@a7a2dc96932430369bd"
            "ef36555082df990ed9bef#egg=marshmallow-annotations"
        )
    ],
    install_requires=[
        # Packages in here should rarely be pinned. This is because these
        # packages (at the specified version) are required for project
        # consuming this library. By pinning to a specific version you are the
        # number of projects that can consume this or forcing them to
        # upgrade/downgrade any dependencies pinned here in their project.
        #
        # Generally packages listed here are pinned to a major version range.
        #
        # e.g.
        # Python FooBar package for foobaring
        # pyfoobar>=1.0, <2.0
        #
        # This will allow for any consuming projects to use this library as
        # long as they have a version of pyfoobar equal to or greater than 1.x
        # and less than 2.x installed.
        "flask>=1.0.2",
        "marshmallow>=2.15.3,<=3.6",
        "marshmallow-annotations",
        "pycryptodome==3.10.1",
    ],
    python_requires=">=3.6",
    package_data={"secoda_common": ["py.typed"]},
)
